import enum
from blang.exceptions import UnexpectedCharacterError
from dataclasses import dataclass
from typing import List
from .matcher import Matcher, Repeat, CommentMatcher, StringMatcher, CharacterMatcher

ALPHA_LOWER = "abcdefghijklmnopqrstuvwxyz"
ALPHA_UPPER = ALPHA_LOWER.upper()
NUMERAL = "0123456789"
PLUSMINUS = "+-"
PERIOD = "."
UNDERSCORE = "_"
SPACE = " "
QUOTE = '"'
EQ = "="
ANY = ALPHA_LOWER + ALPHA_UPPER + NUMERAL + PLUSMINUS + PERIOD + UNDERSCORE + SPACE + EQ


@dataclass
class Token:
    typ: None
    lineno: int
    colno: int
    text: str

    @property
    def type(self):  # why did i call it typ, stupid really
        return self.typ

    def __eq__(self, o):
        if self.typ == o:
            return True
        return False


class TokenSpec(enum.Enum):
    STRING = StringMatcher()
    COMMENT = CommentMatcher("#")
    DEF = Matcher(*"def")
    RETURN = Matcher(*"return")
    EXTERN = Matcher(*"external")
    # EXPORT = Matcher(*"export")
    NEWLINE = Matcher("\n")
    IDENTIFIER = Matcher(
        ALPHA_LOWER + ALPHA_UPPER + UNDERSCORE,
        Repeat(ALPHA_LOWER + ALPHA_UPPER + NUMERAL + UNDERSCORE, min=0),
    )
    WHITESPACE = Matcher(Repeat(" \t", 1))
    LPAREN = Matcher("(")
    RPAREN = Matcher(")")
    LBRACE = Matcher("{")
    RBRACE = Matcher("}")
    LSQBRACKET = Matcher("[")
    RSQBRACKET = Matcher("]")

    LESS_THAN = Matcher(*"<")
    MORE_THAN = Matcher(*">")
    LESS_THAN_EQ = Matcher(*"<=")
    MORE_THAN_EQ = Matcher(*">=")
    EQUAL = Matcher(*"==")
    NOT_EQ = Matcher(*"!=")
    COLON = Matcher(":")
    TERMINATOR = Matcher(";")
    MINUS = Matcher("-")
    PLUS = Matcher("+")
    DIVIDE = Matcher("/")
    MODULO = Matcher("%")
    ASTRISK = Matcher("*")
    COMMA = Matcher(",")
    DOTDOT = Matcher(*"..")
    DOT = Matcher(".")
    ELLIPSIS = Matcher(*"...")

    BAR = Matcher("|")

    BIT_AND = Matcher(*"bit-and")
    BIT_OR = Matcher(*"bit-or")
    BIT_NOT = Matcher(*"bit-not")
    BIT_XOR = Matcher(*"bit-xor")
    BIT_LSHIFT = Matcher(*"bit-lshift")
    BIT_RSHIFT = Matcher(*"bit-rshift")

    CHARACTER = CharacterMatcher()

    # STRING = Matcher(QUOTE, Repeat(ANY), QUOTE)
    FLOAT = Matcher(Repeat(NUMERAL, min=0), PERIOD, Repeat(NUMERAL, min=1))
    INTEGER = Matcher(Repeat(NUMERAL, min=1))
    ASSIGN = Matcher("=")

    AND = Matcher(*"and")
    OR = Matcher(*"or")
    NOT = Matcher(*"not")
    TRUE = Matcher(*"true")
    FALSE = Matcher(*"false")

    IF = Matcher(*"if")
    ELSE = Matcher(*"else")

    WHILE = Matcher(*"while")
    BREAK = Matcher(*"break")
    CONTINUE = Matcher(*"continue")
    FOR = Matcher(*"for")
    AS = Matcher(*"as")

    IMPORT = Matcher(*"import")
    ####
    ## Types
    ##
    U8 = Matcher(*"u8")
    U16 = Matcher(*"u16")
    U32 = Matcher(*"u32")
    U64 = Matcher(*"u64")

    I8 = Matcher(*"i8")
    I16 = Matcher(*"i16")
    I32 = Matcher(*"i32")
    I64 = Matcher(*"i64")

    F64 = Matcher(*"f64")
    BOOL = Matcher(*"bool")
    ####

    @classmethod
    def all_matchers(cls):
        return [t.value for t in cls]

    @classmethod
    def tokenise(cls, input_text) -> List[Token]:
        for matcher in cls.all_matchers():
            matcher.reset()
        p = 0
        line_number = 1
        col_number = 1

        active = set({t for t in cls})
        succeeded = []

        # Append a special \0 to force that there is no active matcher
        # at the end of the string.
        input_text += "\0"
        fed = 0
        while True:
            c = input_text[p]
            p += 1
            fed += 1
            col_number += 1

            # Feed each active matcher with the character
            # and record any completed
            for token in list(active):
                token_matcher = token.value
                match token_matcher.feed(c):
                    case Matcher.FeedResult.CONTINUE:
                        pass
                    case Matcher.FeedResult.FAIL:
                        active.remove(token)
                    case _:
                        succeeded.append(token)
                        active.remove(token)

            if len(active) == 0:
                if len(succeeded) > 0:
                    choice = sorted(
                        succeeded, key=lambda t: t.value.eaten_count, reverse=True
                    )[0]
                    token = Token(
                        typ=choice,
                        colno=col_number,
                        lineno=line_number,
                        text=choice.value.content,
                    )
                    ret = fed - choice.value.eaten_count
                    p -= ret
                    col_number -= ret
                else:
                    if p == len(input_text):
                        break
                    raise UnexpectedCharacterError(c, line_number, col_number - 1)

                # Track the location in the input
                if token.typ == cls.NEWLINE:
                    line_number += 1
                    col_number = 1

                # Reset matchers and status
                for matcher in cls.all_matchers():
                    matcher.reset()

                active = set({t for t in cls})
                succeeded = []
                fed = 0

                if token.typ not in (
                    # TokenSpec.WHITESPACE,
                    # TokenSpec.NEWLINE,
                    TokenSpec.COMMENT,
                ):
                    yield token

            # Don't process the added \0
            if c == "\0":  # p == len(input_text) - 1:
                break



import pytest
from blang.tokeniser import TokenSpec
import blang.exceptions


def test_tokeniser_valid():
    test_string = """
    def def_this_thing_def
       _stinks
    """
    expect = [
        TokenSpec.NEWLINE,
        TokenSpec.WHITESPACE,
        TokenSpec.DEF,
        TokenSpec.WHITESPACE,
        TokenSpec.IDENTIFIER,
        TokenSpec.NEWLINE,
        TokenSpec.WHITESPACE,
        TokenSpec.IDENTIFIER,
        TokenSpec.NEWLINE,
        TokenSpec.WHITESPACE,
    ]
    actual = [t.typ for t in TokenSpec.tokenise(test_string)]
    assert actual == expect


def test_tokeniser_invalid():
    teststr = "something\nsomething~"
    with pytest.raises(blang.exceptions.UnexpectedCharacterError) as exceptioninfo:
        _ = list(TokenSpec.tokenise(teststr))
    assert exceptioninfo.value.line == 2
    assert exceptioninfo.value.c == "~"
    assert exceptioninfo.value.col == 10


def test_tokeniser_trick():
    teststr = "def a_def() defdef"
    expect = [
        TokenSpec.DEF,
        TokenSpec.WHITESPACE,
        TokenSpec.IDENTIFIER,
        TokenSpec.LPAREN,
        TokenSpec.RPAREN,
        TokenSpec.WHITESPACE,
        TokenSpec.IDENTIFIER,
    ]
    actual = [t.typ for t in TokenSpec.tokenise(teststr)]
    assert actual == expect


def test_tokeniser_tricks():
    teststr = "u32u64"
    expect = [
        TokenSpec.IDENTIFIER,
    ]
    actual = [t.typ for t in TokenSpec.tokenise(teststr)]
    assert actual == expect


def test_tokeniser_tricks_typs():
    teststr = "u32 u64"
    expect = [
        TokenSpec.U32,
        TokenSpec.WHITESPACE,
        TokenSpec.U64,
    ]
    actual = [t.typ for t in TokenSpec.tokenise(teststr)]
    print([t.name for t in actual])
    assert actual == expect


def test_tokeniser_int():
    test = """123"""
    expect = [TokenSpec.INTEGER]
    actual = [t.typ for t in TokenSpec.tokenise(test)]
    assert actual == expect


def test_tokeniser_neg_int():
    test = """-123"""
    expect = [TokenSpec.MINUS, TokenSpec.INTEGER]
    actual = [t.typ for t in TokenSpec.tokenise(test)]
    assert actual == expect


def test_tokeniser_int_doubleneg():
    test = """--123"""
    expect = [TokenSpec.MINUS, TokenSpec.MINUS, TokenSpec.INTEGER]
    actual = [t.typ for t in TokenSpec.tokenise(test)]
    print(actual)
    assert actual == expect


def test_tokeniser_float():
    test = """1.5"""
    expect = [TokenSpec.FLOAT]
    actual = [t.typ for t in TokenSpec.tokenise(test)]
    print(actual)
    assert actual == expect



import copy
from dataclasses import dataclass
import enum
from .tokeniser import TokenSpec


class ParseError(Exception):
    def __init__(self, message, node):
        self.message = message
        self.node = node

    def __str__(self):
        return f"{self.message}"


@dataclass
class NodeExtra:
    """Extra detail added to a node before compiling but after parsing."""

    precompiled: bool = False
    is_literal: bool = False
    data_type: str | None = None
    potential_size_mismatch: bool = False
    is_lvalue: bool = False
    is_rvalue: bool = False
    produces_value: bool = False
    is_unary: bool = False
    is_binary: bool = False


class Node:
    def __init__(self, type, token=None, children=None, parent=None, tokens=None):
        self.token = token
        self.children = children or []
        self.parent = parent
        self.type = self.typ = type

        self._tokens = tokens
        self._eaten = 0
        self.blang = ""
        self.extra = NodeExtra()

    def eat(
        self,
        token,
        set_leaf=False,
        set_child=False,
        ignore_space=True,
        ignore_linebreak=True,
    ):
        if len(self._tokens) < self._eaten + 1:
            raise ParseError("unexpected end of input.", self)

        # Eat any whitespace before the token
        original_eaten = self._eaten
        ignores = []
        if ignore_linebreak:
            ignores.append(TokenSpec.NEWLINE)
        if ignore_space:
            ignores.append(TokenSpec.WHITESPACE)
        while (
            ignores
            and len(self._tokens) - 1 > self._eaten
            and self._tokens[self._eaten].typ in ignores
        ):
            self._eaten += 1

        if self._tokens[self._eaten] == token:
            if set_leaf:
                self.token = self._tokens[self._eaten]
            if set_child:
                self.children.append(Node(token, token=self._tokens[self._eaten]))
            self._eaten += 1
            return True

        self._eaten = original_eaten  # restore it not to steal the whitespace
        raise ParseError(f"expected {token}", self)

    def eat_child(self, ChildType):
        n = ChildType(self._tokens[self._eaten :])
        if n:
            self._eaten += n._eaten
            self.children.append(n)
            return True
        raise ParseError(f"Expected {ChildType}", self)

    def peek_child(self, ChildType):
        n = ChildType(self._tokens[self._eaten :])
        return n

    @property
    def id(self):
        return f"{self._tokens[0].lineno}_{self._tokens[0].colno}"

    @property
    def indent_count(self):
        indent = 0
        first_none_white = 0
        while self._tokens[first_none_white].typ in (
            TokenSpec.WHITESPACE,
            TokenSpec.NEWLINE,
        ):
            first_none_white += 1

        # count white spaces spaces backwards
        ix = first_none_white - 1
        while self._tokens[ix].typ == TokenSpec.WHITESPACE:
            indent += len(self._tokens[ix].text)
            ix -= 1

        return indent


def maybe(method):
    def wrapper(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except ParseError:
            return False

    return wrapper


class NodeType(enum.StrEnum):
    UNKNOWN = enum.auto()
    BASE_TYPE = enum.auto()
    REF_TYPE = enum.auto()
    IDENTIFIER_REF = enum.auto()
    ARRAY = enum.auto()
    IDENTIFIER = enum.auto()
    TYPED_IDENTIFER = enum.auto()
    DECLARATION = enum.auto()
    ASSIGNMENT = enum.auto()
    INTEGER = enum.auto()
    FLOAT = enum.auto()
    RETURN = enum.auto()
    FUNC_CALL = enum.auto()
    PARAMETER_LIST = enum.auto()
    FUNC_DEF = enum.auto()
    FUNC_DECL = enum.auto()
    BLOCK = enum.auto()
    CAPTURED_EXPRESSION = enum.auto()
    DE_REF = enum.auto()
    ARRAY_ITEM = enum.auto()
    BOOLEAN = enum.auto()
    ADDITIVE = enum.auto()
    TERM = enum.auto()
    RELATIONAL = enum.auto()
    LOGIC_AND = enum.auto()
    LOGIC_OR = enum.auto()
    NEGATED_RELATIONAL = enum.auto()
    IF_STATEMENT = enum.auto()
    WHILE_LOOP = enum.auto()
    BREAK = enum.auto()
    CONTINUE = enum.auto()
    RANGE = enum.auto()
    FOR_ARRAY_LOOP = enum.auto()
    FOR_RANGE_LOOP = enum.auto()
    MODULE = enum.auto()
    SQUELCH = enum.auto()
    STRING = enum.auto()
    CHARACTER = enum.auto()
    TERMINATOR = enum.auto()
    IMPORT = enum.auto()
    ELLIPSIS = enum.auto()
    BITWISE_OP = enum.auto()


def parser(type: NodeType):
    def parser(method):
        def wrapper(tokens):
            try:
                prototype_node = Node(type, tokens=tokens)
                created_node = method(prototype_node)
                if created_node:
                    # if the created node is not the prototype, maybe as we just want a child,
                    # then we need to stil consider the tokens eaten by the prototype as eatend by
                    # the chosen node
                    created_node._eaten = prototype_node._eaten
                return created_node
            except ParseError:
                return None

        return wrapper

    return parser


def print_tree(node, indent=1):
    if not node:
        print(">>>???<<<")
    else:
        s = "".join(
            (
                " " * (indent - 1),
                "-",
                str(node.typ),
                "  =",
                node.token.text if node.token else "XXXX",
            )
        )
        l = len(s)
        print(s, (70 - l) * " ", "line ", node.token.lineno if node.token else "????")
        for c in node.children:
            print_tree(c, indent + 3)


def OneOf(*args):
    def _OneOf(tokens):
        for Possible in args:
            node = Possible(tokens)
            if node:
                return node
        return None

    return _OneOf


@parser(NodeType.BASE_TYPE)
def BaseType(node):
    for basetype in [
        TokenSpec.U8,
        TokenSpec.U16,
        TokenSpec.U32,
        TokenSpec.U64,
        TokenSpec.I8,
        TokenSpec.I16,
        TokenSpec.I32,
        TokenSpec.I64,
        TokenSpec.F64,
        TokenSpec.BOOL,
    ]:
        if maybe(node.eat)(basetype, set_leaf=True):
            node.blang = node.token.text
            return node
    return None


@parser(NodeType.REF_TYPE)
def RefType(node):
    node.eat(TokenSpec.LESS_THAN)

    if not maybe(node.eat_child)(RefType) and not maybe(node.eat_child)(BaseType):
        return None
    node.eat(TokenSpec.MORE_THAN)
    node.blang = f"<{node.children[0].blang}>"
    return node


@parser(NodeType.IDENTIFIER_REF)
def IdentifierRef(node):
    node.eat(TokenSpec.MORE_THAN)
    node.eat_child(OneOf(ArrayItem, Identifier))
    node.eat(TokenSpec.LESS_THAN)
    node.blang = f">{node.children[0].blang}<"
    return node


@parser(NodeType.CHARACTER)
def Character(node):
    node.eat(TokenSpec.CHARACTER, set_leaf=True)
    node.blang = f"'{node.token.text}'"
    return node


@parser(NodeType.TERMINATOR)
def Terminator(node):
    node.eat(TokenSpec.TERMINATOR, set_leaf=True)
    node.blang = "';'"
    return node


@parser(NodeType.ARRAY)
def Array(node):
    node.eat_child(OneOf(BaseType, RefType))
    node.eat(TokenSpec.LSQBRACKET)
    node.eat_child(Integer)
    node.eat(TokenSpec.RSQBRACKET)
    node.blang = f"{node.children[0].blang}[{node.children[1].blang}]"
    return node


Type = OneOf(Array, BaseType, RefType)  # order matters


@parser(NodeType.SQUELCH)
def Squelch(node):
    node.eat_child(Type)
    node.eat(TokenSpec.BAR)
    node.eat_child(Expr)
    node.eat(TokenSpec.BAR)
    node.blang = f"{node.children[0].blang}|{node.children[1].blang}|"
    return node


@parser(NodeType.IDENTIFIER)
def Identifier(node):
    node.eat(TokenSpec.IDENTIFIER, set_leaf=True)
    node.blang = node.token.text
    node.blang = f"{node.token.text}"
    return node


@parser(NodeType.TYPED_IDENTIFER)
def TypedIdentifier(node):
    node.eat_child(Identifier)
    node.eat(TokenSpec.COLON)
    node.eat_child(Type)
    node.blang = node.children[0].blang + ": " + node.children[1].blang
    return node


@parser(NodeType.DECLARATION)
def Declaration(node):
    is_external = maybe(node.eat)(TokenSpec.EXTERN, set_child=True)
    node.eat_child(TypedIdentifier)
    node.blang = node.children[0].blang
    if not is_external and maybe(node.eat)(TokenSpec.ASSIGN):
        node.eat_child(ExprOrString)
        node.blang += " = " + node.children[-1].blang
    return node


@parser(NodeType.DE_REF)
def DeRef(node):
    node.eat(TokenSpec.LESS_THAN)
    node.eat_child(Additive)
    node.eat(TokenSpec.MORE_THAN)

    node.blang = f"<{node.children[0].blang}>"
    return node


@parser(NodeType.ARRAY_ITEM)
def ArrayItem(node):
    node.eat_child(Identifier)
    node.eat(TokenSpec.LSQBRACKET)
    node.eat_child(Additive)
    node.eat(TokenSpec.RSQBRACKET)
    node.blang = f"{node.children[0].blang}[{node.children[1].blang}]"
    return node


LVal = OneOf(DeRef, ArrayItem, Identifier)  # order matters here


@parser(NodeType.ASSIGNMENT)
def Assignment(node):
    node.eat_child(LVal)
    node.eat(TokenSpec.ASSIGN)
    node.eat_child(ExprOrString)
    node.blang = f"{node.children[0].blang} = {node.children[1].blang}"
    return node


@parser(NodeType.INTEGER)
def Integer(node):
    is_negative = maybe(node.eat)(TokenSpec.MINUS)
    node.eat(TokenSpec.INTEGER, set_leaf=True)
    if is_negative:
        node.token.text = f"-{node.token.text}"
    node.blang = node.token.text
    return node


@parser(NodeType.FLOAT)
def Float(node):
    node.eat(TokenSpec.FLOAT, set_leaf=True)
    node.blang = node.token.text
    return node


Number = OneOf(Integer, Float)


@parser(NodeType.STRING)
def String(node):
    node.eat(TokenSpec.STRING, set_leaf=True)
    node.blang = f'"{node.token.text}"'
    return node


@parser(NodeType.RETURN)
def Return(node):
    node.eat(TokenSpec.RETURN)
    node.blang = "return"
    if maybe(node.eat_child)(Expr):
        node.blang += " " + node.children[0].blang
    return node


@parser(NodeType.FUNC_CALL)
def FuncCall(node):
    node.eat_child(Identifier)
    node.eat(TokenSpec.LPAREN)
    node.blang = node.children[0].blang + "("
    while True:
        if not maybe(node.eat_child)(Expr):
            break
        node.blang += node.children[-1].blang
        if not maybe(node.eat)(TokenSpec.COMMA):
            break
        node.blang += ", "
    node.eat(TokenSpec.RPAREN)
    node.blang += ")"
    return node


@parser(NodeType.ELLIPSIS)
def Ellipsis(node):
    node.eat(TokenSpec.ELLIPSIS, set_leaf=True)
    node.blang = "..."
    return node


@parser(NodeType.PARAMETER_LIST)
def ParameterList(node):
    node.eat(TokenSpec.LPAREN)
    node.blang = "("
    while True:
        if maybe(node.eat_child)(Ellipsis):
            node.blang += "..."
            break

        if not maybe(node.eat_child)(TypedIdentifier):
            break
        node.blang += " " + node.children[-1].blang
        if not maybe(node.eat)(TokenSpec.COMMA):
            break
        node.blang += ","
    node.eat(TokenSpec.RPAREN)
    node.blang += ")"

    return node


@parser(NodeType.FUNC_DEF)
def FuncDef(node):
    node.eat(TokenSpec.DEF)
    node.eat_child(Identifier)
    node.eat_child(ParameterList)
    node.eat(TokenSpec.COLON)
    node.eat_child(Type)
    node.eat_child(Block)
    node.blang = (
        "def "
        + node.children[0].blang
        + node.children[1].blang
        + ": "
        + node.children[2].blang
    )
    return node


@parser(NodeType.FUNC_DECL)
def FuncDecl(node):
    node.eat(TokenSpec.EXTERN)
    node.eat(TokenSpec.DEF)
    node.eat_child(Identifier)
    node.eat_child(ParameterList)
    if maybe(node.eat)(TokenSpec.COLON):
        node.eat_child(Type)
    node.blang = (
        "extern def "
        + node.children[0].blang
        + node.children[1].blang
        + ": "
        + node.children[2].blang
    )
    return node


@parser(NodeType.BLOCK)
def BracedBlock(node):
    node.eat(TokenSpec.LBRACE)
    while True:
        c = maybe(node.eat_child)(Statement)
        if not c:
            break
    node.eat(TokenSpec.RBRACE)
    return node


@parser(NodeType.BLOCK)
def IndentedBlock(node):
    node.eat(TokenSpec.COLON)
    node.eat(TokenSpec.NEWLINE, ignore_space=True, ignore_linebreak=False)
    # now whatever indent comes up, use it to group block
    node.eat_child(Statement)
    indent = node.children[-1].indent_count

    while next := node.peek_child(Statement):
        if next.indent_count == indent:
            node.eat_child(Statement)
        else:
            break

    return node


Block = OneOf(BracedBlock, IndentedBlock)


@parser(NodeType.CAPTURED_EXPRESSION)
def CapturedExpression(node):
    node.eat(TokenSpec.LPAREN)
    node.eat_child(Expr)
    node.eat(TokenSpec.RPAREN)
    node.blang = "( " + node.children[0].blang + ")"
    return node.children[0]


@parser(NodeType.BOOLEAN)
def Boolean(node):
    if maybe(node.eat)(TokenSpec.TRUE, set_leaf=True) or maybe(node.eat)(
        TokenSpec.FALSE, set_leaf=True
    ):
        node.blang = node.token.text
        return node
    return None


Factor = OneOf(
    FuncCall,
    Number,
    Character,
    String,
    Boolean,
    ArrayItem,  # oh no, order matters :-(
    Identifier,
    IdentifierRef,
    CapturedExpression,
    DeRef,
    Squelch,
)


@parser(NodeType.ADDITIVE)
def Additive(node):
    node.eat_child(Term)
    node.blang = node.children[0].blang
    while maybe(node.eat)(TokenSpec.MINUS, set_leaf=True) or maybe(node.eat)(
        TokenSpec.PLUS, set_leaf=True
    ):
        node.blang += " " + node.token.text + " "
        node.eat_child(Term)
        node.blang += node.children[-1].blang
        left = copy.deepcopy(node)
        node.children = []
        node.children.append(left)

    if len(node.children) == 1:
        return node.children[0]
    return node


@parser(NodeType.TERM)
def Term(node):
    node.eat_child(Factor)
    node.blang = node.children[0].blang
    while (
        maybe(node.eat)(TokenSpec.ASTRISK, set_leaf=True)
        or maybe(node.eat)(TokenSpec.DIVIDE, set_leaf=True)
        or maybe(node.eat)(TokenSpec.MODULO, set_leaf=True)
    ):
        node.eat_child(Factor)
        node.blang += node.token.text
        node.blang += node.children[-1].blang
        left = copy.deepcopy(node)
        node.children = []
        node.children.append(left)

    if len(node.children) == 1:
        return node.children[0]
    return node


@parser(NodeType.BITWISE_OP)  # fallback to additive
def BitwiseOp(node):
    unary = False
    node.blang = ""
    if maybe(node.eat)(TokenSpec.BIT_NOT, set_leaf=True):
        node.blang += node.token.text + " "
        unary = True

    node.eat_child(Additive)
    node.blang + node.children[0].blang
    if unary:
        return node

    if (
        maybe(node.eat)(TokenSpec.BIT_OR, set_leaf=True)
        or maybe(node.eat)(TokenSpec.BIT_AND, set_leaf=True)
        or maybe(node.eat)(TokenSpec.BIT_XOR, set_leaf=True)
        or maybe(node.eat)(TokenSpec.BIT_LSHIFT, set_leaf=True)
        or maybe(node.eat)(TokenSpec.BIT_RSHIFT, set_leaf=True)
    ):
        node.eat_child(Additive)
        node.blang += f" {node.token.text} " + node.children[1].blang
        return node
    return node.children[0]


@parser(NodeType.RELATIONAL)
def Relational(node):
    node.eat_child(BitwiseOp)
    node.blang = node.children[0].blang
    if (
        maybe(node.eat)(TokenSpec.MORE_THAN, set_leaf=True)
        or maybe(node.eat)(TokenSpec.LESS_THAN, set_leaf=True)
        or maybe(node.eat)(TokenSpec.MORE_THAN_EQ, set_leaf=True)
        or maybe(node.eat)(TokenSpec.LESS_THAN_EQ, set_leaf=True)
        or maybe(node.eat)(TokenSpec.EQUAL, set_leaf=True)
        or maybe(node.eat)(TokenSpec.NOT_EQ, set_leaf=True)
    ):
        node.eat_child(BitwiseOp)
        node.blang += " " + node.token.text + node.children[1].blang
        return node
    return node.children[0]


@parser(NodeType.LOGIC_AND)
def LogicAnd(node):
    node.eat_child(OneOf(Relational, NegatedRelational))
    node.blang = node.children[0].blang
    if maybe(node.eat)(TokenSpec.AND):
        node.eat_child(OneOf(Relational, NegatedRelational))
        node.blang += " and " + node.children[1].blang
        return node
    return node.children[0]


@parser(NodeType.LOGIC_OR)
def LogicOr(node):
    node.eat_child(LogicAnd)
    node.blang = node.children[0].blang
    if maybe(node.eat)(TokenSpec.OR):
        node.eat_child(LogicAnd)
        node.blang += " or " + node.children[1].blang
        return node
    return node.children[0]


@parser(NodeType.NEGATED_RELATIONAL)
def NegatedRelational(node):
    node.eat(TokenSpec.NOT)
    node.eat_child(Relational)
    node.blang = "not " + node.children[0].blang
    return node


@parser(NodeType.IF_STATEMENT)
def IfStatement(node):
    node.eat(TokenSpec.IF)
    node.eat_child(Expr)
    node.eat_child(Block)

    node.blang = "if " + node.children[0].blang + " (" + node.children[1].blang + ")"
    if maybe(node.eat)(TokenSpec.ELSE):
        node.eat_child(Block)
    return node


@parser(NodeType.WHILE_LOOP)
def WhileLoop(node):
    node.eat(TokenSpec.WHILE)
    node.eat_child(Expr)
    node.eat_child(Block)
    node.blang = "while " + node.children[0].blang

    return node


@parser(NodeType.BREAK)
def Break(node):
    node.eat(TokenSpec.BREAK)
    node.blang = "break"
    return node


@parser(NodeType.CONTINUE)
def Continue(node):
    node.eat(TokenSpec.CONTINUE)
    node.blang = "continue"
    return node


@parser(NodeType.FOR_ARRAY_LOOP)
def ForArrayLoop(node):
    node.eat(TokenSpec.FOR)
    node.eat_child(Identifier)
    node.eat(TokenSpec.AS)
    node.eat_child(Identifier)
    node.eat(TokenSpec.COMMA)
    node.eat_child(Identifier)
    node.eat_child(Block)
    node.blang = f"for {node.children[0].blang} as {node.children[1].blang}, {node.children[2].blang}"
    return node


@parser(NodeType.FOR_RANGE_LOOP)
def ForRangeLoop(node):
    node.eat(TokenSpec.FOR)
    node.eat_child(Range)
    node.eat(TokenSpec.AS)
    node.eat_child(TypedIdentifier)
    node.eat_child(Block)
    node.blang = f"for {node.children[0].blang} as {node.children[1].blang}"

    return node


@parser(NodeType.RANGE)
def Range(node):
    node.eat_child(Expr)
    node.eat(TokenSpec.DOTDOT)
    node.eat_child(Expr)
    if maybe(node.eat)(TokenSpec.COLON):
        node.eat_child(Expr)
    node.blang = f"{node.children[0].blang}..{node.children[1].blang}" + (
        f": {node.children[2].blang}" if len(node.children) > 2 else ""
    )
    return node


ForLoop = OneOf(ForArrayLoop, ForRangeLoop)

Expr = LogicOr  # Additive

ExprOrString = OneOf(Expr, String)

Statement = OneOf(
    FuncCall,
    # FuncDef,
    IfStatement,
    WhileLoop,
    ForLoop,
    Assignment,
    Declaration,
    Return,
    Break,
    Continue,
    Block,
    Terminator,  # a no op
)


@parser(NodeType.IMPORT)
def Import(node):
    node.eat(TokenSpec.IMPORT)
    node.blang = "import "
    # dot separated path
    if maybe(node.eat)(TokenSpec.DOT, set_child=True):
        node.blang += "."

    node.eat_child(Identifier)
    node.blang += node.children[-1].token.text
    while maybe(node.eat)(TokenSpec.DOT):
        node.eat_child(Identifier)
        node.blang += "." + node.children[-1].token.text

    return node


DecOrDef = OneOf(FuncDef, FuncDecl, Declaration, Terminator, Import)


@parser(NodeType.MODULE)
def Module(node):
    while maybe(node.eat_child)(DecOrDef):
        continue
    # ignore white space at the end too
    while maybe(node.eat)(TokenSpec.WHITESPACE) or maybe(node.eat)(TokenSpec.NEWLINE):
        continue
    if node._eaten != len(node._tokens):
        print(f"UNEXPECTED TOKEN '{node._tokens[node._eaten].text}'")
        print("_tokens=", node._tokens)
        print("type=", node.type)
        print("Line number:", node._tokens[node._eaten].lineno)
        raise ParseError(f"unexpected token {node._tokens[node._eaten]}", node)
    return node



from blang.parser import (
    TokenSpec,
    Block,
    Assignment,
    Additive,
    Factor,
    Type,
    ParameterList,
    FuncDef,
    TypedIdentifier,
    String,
    Return,
    Declaration,
    Identifier,
    ArrayItem,
    Range,
    Squelch,
    print_tree,
)


def test_block():
    s = """{
      fisher : u32 
      fisher = fun(45.98+19, 12,other(99))
      print(fisher)
      return 43
    }"""

    tokens = list(TokenSpec.tokenise(s))
    for t in tokens:
        print(t.typ.name)
    print()
    t = Block(tokens)
    print_tree(t)
    # assert False
    assert t._eaten == len(tokens)


def test_assign():
    s = "fisher = fun(45.98+19, 12,other(99))"

    tokens = list(TokenSpec.tokenise(s))
    for t in tokens:
        print(t.typ.name)
    print()
    t = Assignment(tokens)
    print_tree(t)
    # assert False
    assert t._eaten == len(tokens)


def test_assign_ref():
    s = "<fisher> = 10"

    tokens = list(TokenSpec.tokenise(s))
    for t in tokens:
        print(t.typ.name)
    print()
    t = Assignment(tokens)
    print_tree(t)
    # assert False
    assert t._eaten == len(tokens)


def test_term():
    s = "45 + 98-12 * (75.2 - 12)"

    tokens = list(TokenSpec.tokenise(s))
    for t in tokens:
        print(t.typ.name)
    t = Additive(tokens)
    print_tree(t)
    assert t._eaten == len(tokens)


def test_number_factor():
    s = "(654.978)"

    tokens = list(TokenSpec.tokenise(s))
    for t in tokens:
        print(t.typ.name)
    t = Factor(tokens)
    print_tree(t)
    # assert False
    assert t._eaten == len(tokens)
    assert t.token.typ == TokenSpec.FLOAT
    assert float(t.token.text) == 654.978


def test_types():
    s = "u32"
    tokens = list(TokenSpec.tokenise(s))
    t = Type(tokens)
    assert t._eaten == 1
    # assert isinstance(t, BaseType)
    assert t.token.typ == TokenSpec.U32
    # print_tree(t)
    # assert False


def test_base_type():
    s = "u32"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = Type(tokens)
    print_tree(t)
    assert t._eaten == 1
    # assert isinstance(t, BaseType)
    # assert False


def test_ref_type():
    s = "<<u32>>"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = Type(tokens)
    print_tree(t)
    assert t._eaten == 5
    # assert isinstance(t, RefType)
    # assert False


def test_paramlist():
    s = "(fish:f64, face:<i8>) {}"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = ParameterList(tokens)
    assert t
    assert t._eaten == 12
    print(t.children)
    print_tree(t)
    # assert False


def test_func():
    #
    s = """
    def MyFun(fish:f64, face:<i8>):u8{
    
       a: u8 = 9
       b: <u8> = >a<
       c: u8 = a;
       <b> = 9
       return c
    }
    """
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = FuncDef(tokens)
    assert t
    assert t._eaten == len(tokens) - 2
    print(t.children)
    print_tree(t)


def test_typed_ident():
    s = "fish: f64"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = TypedIdentifier(tokens)
    assert t
    assert t._eaten == 4  # there is a space
    print(t.children)
    # assert isinstance(t, TypedIdentifier)
    print_tree(t)
    # assert False


def test_typed_ident_no_type():
    s = "fish"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = TypedIdentifier(tokens)
    assert not t


def test_identifier():
    s = "fish"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = Identifier(tokens)
    assert t
    # assert t._eaten == 3
    print(t)
    print(t.children)
    print_tree(t)
    # assert False


def test_decl():
    s = "fisher :u32 = 99"
    # s = "99"

    tokens = list(TokenSpec.tokenise(s))
    for t in tokens:
        print(t.typ.name)
    print()
    t = Declaration(tokens)
    print_tree(t)
    # assert False
    assert t._eaten == len(tokens)


def test_return():
    s = "return 9"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = Return(tokens)
    assert t
    assert t._eaten == len(tokens)
    print(t.children)
    print_tree(t)


def test_types2():
    s = "<<f64>>"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = Type(tokens)
    assert t
    assert t._eaten == len(tokens)
    print(t.children)
    print_tree(t)
    # assert False


def test_array():
    s = "s:u8[10]"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = Declaration(tokens)
    assert t
    assert t._eaten == len(tokens)
    print(t.children)
    print_tree(t)
    # assert False


def test_array_assign():
    s = "q=s[10]"
    tokens = list(TokenSpec.tokenise(s))
    t = Assignment(tokens)
    assert t
    print_tree(t)
    assert t._eaten == len(tokens)


def test_array_assign_lval():
    s = "s[10]=99"
    tokens = list(TokenSpec.tokenise(s))
    t = Assignment(tokens)
    assert t
    print_tree(t)
    assert t._eaten == len(tokens)


def test_array_item():
    s = "s[10]"
    tokens = list(TokenSpec.tokenise(s))
    t = ArrayItem(tokens)
    assert t
    assert t._eaten == len(tokens)
    print_tree(t)


def test_range():
    s = "1..100:5"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = Range(tokens)
    assert t
    assert t._eaten == len(tokens)
    print_tree(t)


def test_range_no_step():
    s = "1..100"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = Range(tokens)
    assert t
    assert t._eaten == len(tokens)
    print_tree(t)


def test_squelch():
    s = "u8|fish|"
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = Squelch(tokens)
    assert t
    assert t._eaten == len(tokens)
    print_tree(t)


def test_str():
    s = '"FISH"'
    tokens = list(TokenSpec.tokenise(s))
    print([t.typ.name for t in tokens])
    t = String(tokens)
    assert t
    assert t._eaten == len(tokens)
    print_tree(t)



import os
from blang.cli import cli
import tempfile
import subprocess
from pathlib import Path
import pytest


TEST_PROGRAM_FILES = list(Path(__file__).parent.glob("test_*.bl"))
TEST_PROGRAM_NAMES = list(
    map(lambda x: str(x.name).replace(".", "_"), TEST_PROGRAM_FILES)
)


@pytest.mark.parametrize("test_program", TEST_PROGRAM_FILES, ids=TEST_PROGRAM_NAMES)
def test_programs(test_program):
    print(f"Testing: {test_program.name}")
    with open(test_program, "r") as f:
        code = f.read()
        code_lines = code.splitlines()
        # Do something with the content
        expected_exit = int(code_lines[0].split("expected_exit_code")[1].split("=")[1])
        print(f"Expecting exit code {expected_exit}")

        ld_flags = []
        for lin in code_lines:
            if "***ld_flag***" in lin:
                ld_flags.append(lin.split("***ld_flag***")[1])

        # extract expected output
        expected_output = []
        i = 1
        block = False
        for l in code_lines:
            if "***expected output***" in l:
                block = True
                start = False
            elif block and not start and "****" in l:
                indent = len(l.split(" ")[0]) + 1
                start = True
            elif block and start and "****" in l:
                block = False
            elif block and start:
                expected_output.append(l[indent:])

        print(f"Expecting stdout \n```\n{'\n'.join(expected_output)}\n```")

        with tempfile.NamedTemporaryFile("w") as f:
            binary = f.name + ".bin"
            try:
                if cli(test_program.absolute(), binary, " ".join(ld_flags), debug=True):
                    assert False, "compile failed"
            except Exception:
                raise
            process = subprocess.run([binary], capture_output=True, text=True)
            result = process.returncode
            output_lines = process.stdout.splitlines()

        assert result == expected_exit

        print(f"Program stdout:\n```\n{process.stdout}```")
        assert output_lines == expected_output, (
            f"Expected:\n{expected_output}\nGot:\n{output_lines}"
        )
        os.remove(binary)




from subprocess import CalledProcessError

import click
import subprocess
from blang.compiler.compiler import (
    compiler,
    CompileError,
    parse_file,
    import_node_to_filepath,
)
from blang.parser import ParseError, NodeType
import tempfile
from pathlib import Path


@click.command
@click.argument("input_file", type=click.Path(exists=True))
@click.argument("output_file", type=click.Path())
@click.option("--debug", is_flag=True, required=False, default=False)
@click.option("--ld_flags", is_flag=False, required=False, default="")
def cli_cmd(input_file, output_file, ld_flags, debug):
    cli(input_file, output_file, ld_flags, debug)


def cli(input_file, output_file, ld_flags="", debug=False):
    """Compiles INPUT_FILES and writes the linked binary to OUTPUT_FILE."""
    object_files = []
    input_files = [input_file]
    module = parse_file(input_file, debug)
    imports = list(filter(lambda x: x.type == NodeType.IMPORT, module.children))
    for import_node in imports:
        input_files.append(import_node_to_filepath(import_node, input_file))

    for asm in (Path(__file__).parent / "compiler").glob("*.asm"):
        cmd = [
            "nasm",
            "-f",
            "elf64",
            str(asm),
            "-o",
            str(asm.with_suffix(".o")),
        ]
        click.echo(">  " + " ".join(cmd))
        subprocess.check_call(cmd)
        object_files.append(str(asm.with_suffix(".o")))

    for file in input_files:
        with tempfile.NamedTemporaryFile("w") as tmp:
            click.echo(f"Compiling {file}")
            try:
                asm = compiler(file, debug)
            except CompileError as e:
                click.echo(f"Compilation error:\n {str(e)}", err=True)
                return 1
            except ParseError as e:
                click.echo(f"Syntax error:\n {str(e)}", err=True)
                return 1
            tmp.write("\n".join(asm or []))
            tmp.flush()
            if not asm:
                print("Problem compiling.")
                return 1
            if debug:
                print("******" * 5)
                print({file})
                for i, l in enumerate(asm or []):
                    print(f"{i + 1}\t{l}")
            cmd = ["nasm", "-f", "elf64", tmp.name, "-o", tmp.name + ".o"]
            click.echo(">" + " ".join(cmd))
            try:
                subprocess.check_call(cmd)
            except CalledProcessError as e:
                print("Problem assembling code.")
                print(e)
                return 1
            object_files.append(tmp.name + ".o")

    click.echo(f"Linking {object_files}")
    cmd = ["ld", *map(str, object_files), *(ld_flags.split()), "-o", output_file]
    click.echo("> " + " ".join(cmd))
    try:
        subprocess.check_call(cmd)
    except CalledProcessError as e:
        print("Problem linking code")
        print(e)
        return 1
    click.echo(f"Executable written to {output_file}")
    return 0


def main():
    try:
        cli_cmd()
    except subprocess.CalledProcessError:
        click.echo("OhNo.", err=True)
        return 1



import enum
import copy
from blang.exceptions import UnexpectedCharacterError
from dataclasses import dataclass
from typing import List

ALPHA_LOWER = "abcdefghijklmnopqrstuvwxyz"
ALPHA_UPPER = ALPHA_LOWER.upper()
NUMERAL = "0123456789"
PLUSMINUS = "+-"
DECIMAL = "."
UNDERSCORE = "_"
SPACE = " "


@dataclass
class Token:
    typ: None
    lineno: int
    colno: int
    text: str


class Repeat:
    def __init__(self, symbol, min=1, max=None):
        self.sym = symbol
        self.min = min
        self.max = max
        self.count = 0

    def hit(self):
        self.count += 1

    @property
    def is_full(self):
        return self.max and self.count == self.max


class Matcher:
    class FeedResult(enum.Enum):
        FAIL = enum.auto()
        CONTINUE = enum.auto()
        DONE = enum.auto()
        DONE_NOT_EATEN = enum.auto()

    def __init__(self, *args):
        # exp is a list of string | list
        # when its a list, its a or operation on all in the list
        # when its a string, its a or operation on the string entries
        self.exp = []
        self.active_exp = []  # working copy of self.exp

        on_repeat = False
        for a in args:
            if isinstance(a, Repeat):  # if its a repeat then join with next
                self.exp.extend([a.sym] * a.min)
                self.exp.append([a])
                on_repeat = True
            else:
                if on_repeat:
                    self.exp[-1].append(a)
                    on_repeat = False
                else:
                    self.exp.append(a)
        self.reset()

    @property
    def active(self):
        return len(self.active_exp) > 0 and self.eaten_count > 0

    @property
    def eaten_count(self):
        return len(self.content)

    def feed(self, c):
        if len(self.active_exp) < 1:
            self.is_failed = True
            return Matcher.FeedResult.FAIL
        next_mach = self.active_exp[0]
        if isinstance(next_mach, str):
            if c in next_mach:
                self.content += c
                self.active_exp.pop(0)
            else:
                self.is_failed = True
                return Matcher.FeedResult.FAIL
        if isinstance(next_mach, list):
            # or on the list, as have a repeater
            repeat = next_mach[0]
            mach = next_mach[1] if len(next_mach) > 1 else None
            if not repeat.is_full and c in repeat.sym:
                self.content += c
                repeat.hit()
                return Matcher.FeedResult.CONTINUE
            elif mach is None:
                self.active_exp.pop(0)
                return Matcher.FeedResult.DONE_NOT_EATEN
            elif c in mach:
                self.content += c
                self.active_exp.pop(0)
            else:
                self.is_failed = True
                return Matcher.FeedResult.FAIL

        if len(self.active_exp) > 0:
            return Matcher.FeedResult.CONTINUE
        else:
            return Matcher.FeedResult.DONE

    def reset(self):
        self.content = ""
        self.active_exp = copy.deepcopy(self.exp)  # coppies the repeats too, nice
        self.is_failed = False


class CommentMatcher:
    """Specialist matcher for catching comments."""

    def __init__(self, comment_char="#"):
        self.comment_char = comment_char
        self.reset()

    @property
    def active(self):
        return len(self.active_exp) > 0 and self.eaten_count > 0

    @property
    def eaten_count(self):
        return len(self.content)

    def feed(self, c):
        if self.is_failed:
            return Matcher.FeedResult.FAIL
        if self.in_comment:
            if c == "\n":
                return Matcher.FeedResult.DONE_NOT_EATEN
            self.content += c
            return Matcher.FeedResult.CONTINUE
        if c == self.comment_char:
            self.in_comment = True
            self.content += c
            return Matcher.FeedResult.CONTINUE
        self.is_failed = True
        return Matcher.FeedResult.FAIL

    def reset(self):
        self.content = ""
        self.in_comment = False
        self.is_failed = False


class StringMatcher:
    """Specialist matcher for catching strings."""

    def __init__(self, str_char='"'):
        self.str_char = str_char
        self.reset()

    @property
    def active(self):
        return len(self.active_exp) > 0 and self.eaten_count > 0

    @property
    def eaten_count(self):
        return len(self.content) + 2

    def feed(self, c):
        if self.is_failed:
            return Matcher.FeedResult.FAIL
        if self.in_string:
            if c == self.str_char:
                return Matcher.FeedResult.DONE_NOT_EATEN
            self.content += c
            return Matcher.FeedResult.CONTINUE
        if c == self.str_char:
            self.in_string = True
            return Matcher.FeedResult.CONTINUE
        self.is_failed = True
        return Matcher.FeedResult.FAIL

    def reset(self):
        self.content = ""
        self.in_string = False
        self.is_failed = False


class CharacterMatcher:
    """Specialist matcher for catching strings."""

    def __init__(self, str_char="'"):
        self.str_char = str_char
        self.reset()

    @property
    def active(self):
        return len(self.active_exp) > 0 and self.eaten_count > 0

    @property
    def eaten_count(self):
        return len(self.content) + 2

    def feed(self, c):
        if self.is_failed:
            return Matcher.FeedResult.FAIL
        if self.in_string:
            if c == self.str_char:
                if len(self.content) != 1:
                    return Matcher.FeedResult.FAIL
                return Matcher.FeedResult.DONE_NOT_EATEN
            self.content += c
            return Matcher.FeedResult.CONTINUE
        if c == self.str_char:
            self.in_string = True
            return Matcher.FeedResult.CONTINUE
        self.is_failed = True
        return Matcher.FeedResult.FAIL

    def reset(self):
        self.content = ""
        self.in_string = False
        self.is_failed = False



from pathlib import Path

from blang.compiler.precompiler import precompile
from blang.parser import Module, NodeType, Node, print_tree
from blang import parser
from blang.tokeniser import TokenSpec
import os

from blang.compiler.types import (
    Literal,
    Register,
    Variable,
    Array,
    Function,
    Context,
    ArgumentRegistersBySize,
    TokenVariableTypeMap,
    VariableType,
    SizeSpecifiers,
    SizeReserves,
    TypeSizes,
    SizeDefiners,
    AddressPointer,
    RBP,
    SignedIntVariableTypes,
    FloatArgumentRegisters,
    LiteralString,
    NodeCompilerValueType,
)


class CompileError(Exception):
    def __init__(self, message, node):
        self.message = message
        self.node = node

    def __str__(self):
        return f"line {self.node._tokens[0].lineno} \n" + self.message


_node_compilers = {}


def node_compiler(type: NodeType):
    def node_compiler(method):
        def _wrap(*args, **kwargs):
            return method(*args, **kwargs)

        _node_compilers[type] = _wrap
        return _wrap

    return node_compiler


def compile(node: Node, context, **kwargs):
    assert node.extra.precompiled, "Compiler bug. Precompiled node expected."
    try:
        asm, registers = _node_compilers[node.type](node, context, **kwargs)

        if asm:
            spacing = max(30 - len(asm[0]), 0)
            # line number - first not whitespace token
            i = 0
            while node._tokens[i].typ in [TokenSpec.WHITESPACE, TokenSpec.NEWLINE]:
                i += 1

            asm[0] += spacing * " " + f";  L{node._tokens[i].lineno}: " + node.blang
    except:  # AttributeError:
        print_tree(node)
        print(node, node.type)
        raise  # CompileError(f"Bad times. {e}", node)
    return asm, registers


def type_to_base_type_and_indirection_count(type) -> (VariableType, int):
    refs = 0
    while type.typ == NodeType.REF_TYPE:
        type = type.children[0]
        refs += 1
    return TokenVariableTypeMap[type.token.typ], refs


def typed_identifier_to_variable(typed_identifier: Node):
    identifier, type = typed_identifier.children
    array_size = 0
    if type.type == NodeType.ARRAY:
        array_size = int(type.children[1].token.text)
        type = type.children[0]

    type, refs = type_to_base_type_and_indirection_count(type)

    if array_size > 0:
        return Array(
            identifier=identifier.token.text,
            type=type,
            indirection_count=refs,
            length=array_size,
        )

    return Variable(identifier=identifier.token.text, type=type, indirection_count=refs)


def collect_functions_from_module(module: Node, context: Context):
    """Extracts function declarations from a module and adds Function objects
    for them to the context.variables. Returns list of added functions."""
    functions = []
    # pass 1 to get function declarations into globals
    for function in list(
        filter(
            lambda x: x.type in (parser.NodeType.FUNC_DEF, parser.NodeType.FUNC_DECL),
            module.children,
        )
    ):
        identifier, parameters, type = function.children[:3]
        type, refs = type_to_base_type_and_indirection_count(type)
        f = Function(
            identifier=identifier.token.text,
            type=type,
            indirection_count=refs,
            parameters=[
                typed_identifier_to_variable(t) if t.children else t
                for t in parameters.children
            ],
            node=function,
            variadic=parameters.children
            and parameters.children[-1].type == NodeType.ELLIPSIS,
        )
        if f.identifier in context.variables:
            raise CompileError(f"Duplicate declaration of '{f.identifier}'", identifier)
        context.globals_[f.identifier] = f
        functions.append(f)
    return functions


def import_node_to_filepath(node, imported_from: Path):
    if node.children[0].type == TokenSpec.DOT:
        filename = imported_from.parent / os.path.join(
            *(n.token.text for n in node.children)
        )
    else:
        filename = Path("/") / os.path.join(*(n.token.text for n in node.children))
    return filename.with_suffix(".bl")


@node_compiler(parser.NodeType.IMPORT)
def compile_import(node, context: Context) -> str:
    asm = []
    filename = import_node_to_filepath(node, context.filename)
    print(filename)
    # open the file, read it, parse it,
    module = parse_file(filename)

    # extract exported declarations, produce a global directive for them and
    # put them into context

    # Collect declarations
    declarations = list(
        filter(lambda x: x.type == parser.NodeType.DECLARATION, module.children)
    )
    for dec in declarations:
        extern = dec.children[0].token and dec.children[0].token.typ == TokenSpec.EXTERN
        var = typed_identifier_to_variable(
            dec.children[0] if not extern else dec.children[1]
        )
        var.location = AddressPointer(var.identifier)
        context.globals_[var.identifier] = var
        asm.append(f"extern {var.identifier}")

    # Collect functions
    functions = collect_functions_from_module(module, context)
    for function in functions:
        asm.append(f"extern {function.identifier} ")

    return asm, []


def collect_children_of_type(parent, type: NodeType):
    nodes = [parent]
    collection = []
    while len(nodes) > 0:
        n = nodes.pop(0)
        for child in n.children:
            nodes.append(child)
        if n.type == type:
            collection.append(n)
    return collection


@node_compiler(parser.NodeType.MODULE)
def compile_module(node, context: Context) -> list[str]:
    asm = []
    # Collect imports
    imports = list(filter(lambda x: x.type == parser.NodeType.IMPORT, node.children))
    for i in imports:
        code, _ = compile(i, context)
        asm.extend(code)

    # Store literals into .rodata
    asm.append("section .rodata")
    # Collect all strings together to keep in .ro
    string_literals = collect_children_of_type(node, NodeType.STRING)
    for n in string_literals:
        if n.token.text in context.string_literals:
            continue
        s_id = f"___s{len(context.string_literals) + 1}"
        context.string_literals[n.token.text] = s_id
        asm.append(f"{s_id}: db  {string_to_nasm(n.token.text)}, 0")

    # Literal floats have to be stored
    float_literals = collect_children_of_type(node, NodeType.FLOAT)
    for flt_node in float_literals:
        if flt_node.token.text in context.float_literals:
            continue
        f_id = f"___f{len(context.float_literals) + 1}"
        asm.append(f"{f_id}: dq {flt_node.token.text}")
        var = Variable(
            identifier=f_id,
            type=VariableType.f64,
            indirection_count=0,
            location=f"[{f_id}]",
        )
        context.float_literals[flt_node.token.text] = var

    # Collect declarations
    declarations = list(
        filter(lambda x: x.type == parser.NodeType.DECLARATION, node.children)
    )
    asm.append("section .data")
    for declaration in filter(lambda d: len(d.children) > 1, declarations):
        code, _ = compile(declaration, context)
        asm.extend(code)
    asm.append("")
    asm.append("section .bss")
    for declaration in filter(lambda d: len(d.children) < 2, declarations):
        code, _ = compile(declaration, context)
        asm.extend(code)
    asm.append(" ")

    # Collect funcs, adds them to context
    functions = collect_functions_from_module(node, context)

    # Compile definitions of the collected functions
    asm.append("section .text")
    for function in functions:
        code, _ = compile(function.node, context)
        asm.extend(code)

    assert len(context.occupied_registers) == 0, (
        f"Taken regs: {context.occupied_registers}\n{'\n'.join(asm)}'"
    )
    return asm, []


def string_to_nasm(s):
    n = [""]
    p = 0
    escaping = False
    while p < len(s):
        c = s[p]
        p += 1
        if escaping and c == "\\":
            n[-1].append("\\")
            escaping = False
        elif escaping and c == "n":
            n.append(ord("\n"))
            n.append("")
            escaping = False
        elif escaping and c == "t":
            n.append(ord("\t"))
            n.append("")
            escaping = False
        elif escaping and c == "0":
            n.append(ord("\0"))
            n.append("")
            escaping = False
        elif c == "\\":
            escaping = True
            continue
        else:
            n[-1] += c
    return ", ".join(map(lambda x: f'"{x}"' if isinstance(x, str) else f"{x}", n))


def test_str_nasm():
    print(string_to_nasm("this\\0that\\nIS\\tagain"))


@node_compiler(parser.NodeType.DECLARATION)
def compile_declaration(node, context: Context):
    extern = node.children[0].token and node.children[0].token.typ == TokenSpec.EXTERN

    var = typed_identifier_to_variable(
        node.children[0] if not extern else node.children[1]
    )
    init = None
    if len(node.children) > (2 if extern else 1):
        if extern:
            raise CompileError("Can't set a intialiser to an external. ", node)
        init = node.children[1]

    if extern or not context.use_stack:
        if context.is_local_var(var.identifier):
            raise CompileError(
                f"Variable '{var.identifier}' is already defined.", node.children[0]
            )

        var.location = AddressPointer(var.identifier)
        context.globals_[var.identifier] = var

        if extern:
            return [
                f"extern {var.identifier}"
            ], []  # todo this is going in the data sect

        if init:
            code, (literal,) = compile(init, context)
            if not isinstance(literal, Literal):
                raise CompileError(
                    f"Can't initialise '{var.identifier}' with non-constant value.",
                    init,
                )
            # todo check sizes match
            dd = SizeDefiners[var.size]
            return [
                f"global {var.identifier}",
                f"{var.identifier}: {dd} {literal.value}",
            ], []

        res = SizeReserves[var.size]
        count = 1 if not isinstance(var, Array) else var.length
        return [f"global {var.identifier}", f"{var.identifier}: {res} {count}"], []
    else:
        if context.is_local_var(var.identifier):
            raise CompileError(
                f"Variable '{var.identifier}' is already defined.", node.children[0]
            )

        count = 1 if not isinstance(var, Array) else var.length
        context.locals_stack_size += count * var.size
        var.location = RBP(-context.locals_stack_size)
        context.variables[var.identifier] = var
        initialise = ()
        if init:
            if init.type == NodeType.STRING:
                if var.type != VariableType.u8 and not isinstance(var, Array):
                    raise CompileError("Only assign strings to u8[] types.", node)
                src_str = init.token.text
                src_id = context.string_literals.get(src_str)

                if not src_id:
                    raise CompileError(
                        f'Compiler bug. Missing string literal "{src_str}" in ro data.',
                        node,
                    )

                str_len = len(src_str)
                if str_len > var.length:
                    raise CompileError(
                        f"'{src_str} doesnt fit into {var.identifier}. Needs {str_len} bytes.",
                        node,
                    )
                initialise = [
                    "cld",
                    f"mov rsi, {src_id}         ; source address",
                    f"lea rdi, {var.location}   ; destination address",
                    f"mov rcx, {str_len}                 ; number of bytes to copy",
                    "rep movsb",
                ]
            else:
                sizespec = SizeSpecifiers[var.size]
                asm, (reg,) = compile(init, context)
                if not isinstance(reg, Literal):  # err hack ot avoid a:8 = u8|9|
                    reg = ensure_is_a_register(asm, reg, context)
                if var.type != reg.type:
                    if isinstance(reg, Literal):
                        reg.type = var.type
                    else:
                        raise CompileError(
                            f"Type mismatch, {var.identifier} is {var.type} while rval({reg}) is {reg.type}",
                            node,
                        )
                initialise = [*asm, f"mov {sizespec} {var.location}, {reg}"]
                context.mark_free_if_reg(reg)

        return [f"; {var.identifier} @ {var.location}", *initialise], []


@node_compiler(NodeType.FUNC_DECL)
def compile_func_decl(node, context: Context):
    identifier = node.children[0]
    return [f"extern {identifier.token.text}"], []


@node_compiler(NodeType.FUNC_DEF)
def compile_func(node, context: Context):
    # we already created the Function object in the globals as part of the module
    # so we can reuse it
    identifier, parameters, type, block = node.children
    function: Function = context.variables.get(identifier.token.text)
    if not function:
        # we could create it here to make sure to support compile this without modile compile
        # assumptions. todo
        ...

        raise NotImplementedError(
            "Need to make funcs compile not relying on module to create  declaration."
        )
    context.locals_stack_size = 0
    context.use_stack = True
    context.current_func = function
    # Set the parameters as locals
    context.new_frame()
    rbp_offset = 16  # todo why this? stack params are broke
    for i, parameter in enumerate(function.parameters):
        if i > 5:
            location = RBP(-rbp_offset)
            rbp_offset += parameter.size
        else:
            location = ArgumentRegistersBySize[parameter.size][i]

        context.variables[parameter.identifier] = Variable(
            parameter.identifier,
            location=location,
            type=parameter.type,
            indirection_count=parameter.indirection_count,
            on_stack=False,
        )

    blk, _ = compile(block, context)
    context.pop_frame()
    context.current_func = None
    return [
        "",
        f"global {function.identifier}",  # if function.exported
        f"{function.identifier}:",
        "push rbp",
        "mov rbp, rsp",
        f"sub rsp, {(context.locals_stack_size // 16 + 1) * 16}",
        *blk,
        f"{function.identifier}___ret:",
        "leave",
        "ret",
    ], []


@node_compiler(NodeType.BLOCK)
def compile_block(node, context: Context):
    asm = []
    context.new_frame()
    for child in node.children:
        print(child.type)
        code, registers = compile(child, context)
        asm.extend(code)
        for reg in registers:
            # some statements such a funccall may have taken a register
            context.mark_free_if_reg(reg)
    context.pop_frame()
    return asm, []


@node_compiler(NodeType.IDENTIFIER_REF)
def compile_ref(node, context: Context):
    asm, [reg] = compile(
        node.children[0], context, compile_type=NodeCompilerValueType.lval
    )
    return asm, [reg]


def ensure_is_a_register(code, maybe_reg, context):
    """Ensures value is in a register by allocating one if needed.

    :param code: List of assembly instructions to append new instructions to
    :param maybe_reg: Register or other value that needs to be in a register
    :param context: Compilation context for register allocation
    :return: Tuple of (assembly instructions, register containing value)
    """
    if isinstance(maybe_reg, Register):
        return maybe_reg
    else:
        reg = context.take_a_register()
        reg.type = maybe_reg.type
        reg.indirection_count = maybe_reg.indirection_count
        code.append(f"mov {reg.full_reg}, {maybe_reg}")
        return reg


@node_compiler(NodeType.DE_REF)
def compile_de_ref(
    node,
    context: Context,
    compile_type: NodeCompilerValueType = NodeCompilerValueType.rval,
):
    # note, this is compiling only a an rval, in assignment this is not run
    code, (ptr,) = compile(node.children[0], context)
    reg = ensure_is_a_register(code, ptr, context)

    if reg.indirection_count < 1:
        raise CompileError("Problem. Can't dereference a non-reference.", node)
    reg.indirection_count -= 1
    asm = [
        *code,
        f"mov {SizeSpecifiers[reg.size]}  {reg}, [{reg.full_reg}]; ;cnt={reg.indirection_count}",
    ]

    return asm, [reg]


@node_compiler(NodeType.ARRAY_ITEM)
def compile_array_item(
    node,
    context: Context,
    compile_type: NodeCompilerValueType = NodeCompilerValueType.rval,
):
    # when assigning, this is lval and we return the address, normal case is rval so return value

    identifier_node, index_expr = node.children
    code, (index,) = compile(index_expr, context)
    identifier = context.variables[identifier_node.token.text]

    reg = context.take_a_register()
    tmp = context.free_registers[0]
    tmp.type = index.type
    tmp.indirection_count = 0
    reg.type = identifier.type
    reg.indirection_count = index.indirection_count
    code = [
        *code,
        f"lea {reg.full_reg}, {identifier}",
        f"xor {tmp.full_reg}, {tmp.full_reg}",  # horid temp to zero upper bits for addition
        f"mov {tmp}, {index}",
        *((f"add {reg.full_reg}, {tmp.full_reg}",) * identifier.size),
    ]

    asm = [*code]
    context.mark_free_if_reg(index)

    if compile_type is NodeCompilerValueType.lval:
        reg.indirection_count = 1
        return asm, [reg]

    # use the same register, drop size
    asm.append(f"mov {SizeSpecifiers[reg.size]}  {reg}, [{reg.full_reg}] ; load it ")
    return asm, [reg]


@node_compiler(NodeType.IDENTIFIER)
def compile_identifier(
    node,
    context: Context,
    compile_type: NodeCompilerValueType = NodeCompilerValueType.rval,
):
    var_name = node.token.text
    var = context.variables.get(var_name)
    if not var:
        raise CompileError("Unknown variable '{var_name}'", node)
    if compile_type is NodeCompilerValueType.lval:
        reg = context.take_a_register()
        reg.type = var.type
        reg.indirection_count = var.indirection_count + 1
        return [f"lea {reg}, {var.location}"], [reg]
    return [], [context.variables[var_name]]


@node_compiler(NodeType.INTEGER)
def compile_integer(node, context: Context):
    # todo literal sizes and types
    return [], [Literal(node.token.text, type=VariableType.u32)]


@node_compiler(NodeType.CHARACTER)
def compile_literal_character(node, context: Context):
    return [], [Literal(str(ord(node.token.text)), type=VariableType.u8)]


@node_compiler(NodeType.FLOAT)
def compile_literal_float(node, context: Context):
    return [], [context.float_literals[node.token.text]]


@node_compiler(NodeType.STRING)
def compile_literal_string(node, context: Context):
    val = string_to_nasm(node.token.text)
    # return [], [Literal(val, type=VariableType.u8, length=len(node.token.text))]
    string_literal = context.string_literals[node.token.text]
    return [], [
        LiteralString(
            "tmpstring",
            type=VariableType.u8,
            length=len(node.token.text),
            node=node,
            location=string_literal,
            value=val,
        )
    ]


@node_compiler(NodeType.ADDITIVE)
def compile_additive(node: Node, context: Context):
    op = "add" if node.token == TokenSpec.PLUS else "sub"
    code_l, (l,) = compile(node.children[0], context)
    reg = ensure_is_a_register(code_l, l, context)
    lval_is_flt = reg.type == VariableType.f64

    code_r, (r,) = compile(node.children[1], context)
    if (not isinstance(r, Literal) or lval_is_flt) and r.size != reg.size:
        raise CompileError(
            f"Size type mismatch. {node.children[1].blang} is {r.type} = {r.size} bytes,  {node.children[0].blang} is {l.type} = {l.size} bytes. May need to squelch.",
            node,
        )
    if r.type != reg.type or (r.indirection_count > 1 != reg.indirection_count > 1):
        print(
            f"WARNING: L{node.token.lineno} Type mismatch in addition. Sizes match so proceeding. May be bad."
        )
        print(f"       Lval {reg.type}, Rval {r.type}")

    context.mark_free_if_reg(r)

    if lval_is_flt:
        # add two floats together using xmm
        f_reg = context.take_a_float_register()
        r = ensure_is_a_register(code_r, r, context)
        asm = [
            *code_l,
            *code_r,
            "sub rsp, 16",
            #
            f"mov qword [rsp], {reg}",
            f"mov qword [rsp-8], {r}",  #
            f"movsd {f_reg}, [rsp]",
            f"{op}sd {f_reg}, [rsp-8]",
            f"movsd [rsp], {f_reg}",
            f"mov qword {reg}, [rsp]",
            # "movsd xmm0, [rsp]",  # todo cheat for printf
            #
            # "xor rax,rax",
            # "mov rax, 1",
            "add rsp, 16",
        ]
        context.free_a_float_register(f_reg)
        context.mark_free_if_reg(r)
        return asm, [reg]

    ref_muliplier = 1
    if reg.indirection_count:
        reg.indirection_count -= 1
        ref_muliplier = reg.size
        reg.indirection_count += 1

    return [
        *code_l,
        *code_r,  #
        *(f"{op} {reg}, {r}" for _ in range(ref_muliplier)),
    ], [reg]


@node_compiler(NodeType.TERM)
def compile_term(node: Node, context: Context):
    code_l, (l,) = compile(node.children[0], context)

    code_r, (r,) = compile(node.children[1], context)
    if not isinstance(r, Literal) and r.type != l.type:
        raise CompileError(
            f"Size mismatch. {r} is {r.size} bytes {r.type}, {l} is {l.size} bytes {l.type}. May need to squelch.",
            node,
        )
    if r.type != l.type or (r.indirection_count > 1 != l.indirection_count > 1):
        print(
            "WARNING: Type mismatch in multiplication. Sizes match so proceeding. May be bad."
        )

    if r.indirection_count or l.indirection_count:  # either is indirection
        raise CompileError("Can't multiply or divide with a pointer.", node)

    reg = ensure_is_a_register(code_l, l, context)

    asm = [
        *code_l,
        *code_r,
    ]

    context.mark_free_if_reg(r)

    # if r is not a reg, make it so because mul needs registers
    if not isinstance(r, Register):
        rcx = Register("rcx")
        rcx.type = r.type
        asm.append(f"mov {rcx}, {r}")
        r = rcx

    rax = Register("rax")
    rax.type = l.type
    rdx = Register("rdx")
    rdx.type = l.type
    is_signed = reg.type in [
        VariableType.i8,
        VariableType.i16,
        VariableType.i32,
        VariableType.i64,
    ]
    signed_extend = {8: "cqo", 4: "cdq", 2: "cwd", 1: "cbw"}
    is_float_op = l.type == VariableType.f64

    # result going into reg, which holds lval now and is a register
    # rval is in r, which is a register

    match node.token:
        case TokenSpec.ASTRISK:
            if is_float_op:
                return [
                    *asm,
                    f"mov qword [rsp-8], {reg}",
                    "movsd xmm0, [rsp-8]",
                    f"mov qword [rsp-8], {r}",
                    "movsd xmm1, [rsp-8]",
                    "mulsd xmm0, xmm1",
                    "movsd [rsp-8], xmm0",
                    f"mov qword {reg}, [rsp-8]",
                ], [reg]
            else:
                return [
                    *asm,  #
                    f"mov {rax}, {l}",
                    f"{'mul' if not is_signed else 'imul'} {r.sizespec} {r}",
                    f"mov {reg}, {rax}",
                ], [reg]
        case TokenSpec.DIVIDE:
            if is_float_op:
                return [
                    *asm,
                    f"mov qword [rsp-8], {reg}",
                    "movsd xmm0, [rsp-8]",
                    f"mov qword [rsp-8], {r}",
                    "movsd xmm1, [rsp-8]",
                    "divsd xmm0, xmm1",
                    "movsd [rsp-8], xmm0",
                    f"mov qword {reg}, [rsp-8]",
                ], [reg]
            else:
                return [
                    *asm,  #
                    f"mov {rax}, {l}",
                    "xor rdx, rdx" if not is_signed else signed_extend[rax.size],
                    f"{'div' if not is_signed else 'idiv'} {r.sizespec} {r}",
                    f"mov {reg}, {rax}",
                ], [reg]

        case TokenSpec.MODULO:
            if is_float_op:
                return [
                    *asm,
                    f"mov qword [rsp-8], {r}",
                    "fld qword [rsp-8]",
                    f"mov qword [rsp-8], {reg}",
                    "fld qword [rsp-8]",
                    "fprem",
                    "fstp st1",
                    "fstp qword [rsp-8]",
                    f"mov qword {reg}, [rsp-8]",
                ], [reg]
            else:
                return [
                    *asm,  #
                    f"mov {rax}, {l}",
                    "xor rdx, rdx" if not is_signed else signed_extend[rax.size],
                    f"{'div' if not is_signed else 'idiv'} {r.sizespec} {r}",
                    f"mov {reg}, {rdx}",
                ], [reg]
        case _:
            raise CompileError("Compiler bug.", node)


@node_compiler(NodeType.ASSIGNMENT)
def compile_assignment(node, context: Context):
    # compile what will be assigned to it
    asm, (reg,) = compile(node.children[1], context)

    # put it into a register as some ops cant work off memory
    if not isinstance(
        reg, Literal
    ):  # this is a hack to allow a:u8=9 otherwise 9 is a u32 and needs squelching
        reg = ensure_is_a_register(asm, reg, context)

    if node.children[0].type == NodeType.DE_REF:
        decount = 1
        id_node = node.children[0].children[0]
        while id_node.type == NodeType.DE_REF:
            decount += 1
            id_node = id_node.children[0]
        get_addr_asm, (lval,) = compile(id_node, context)

        if lval.indirection_count < decount:
            raise CompileError("Tried to deref too many times in lval.", node)

        lval.indirection_count -= decount
        sizespec = SizeSpecifiers[lval.size]
        lval.indirection_count += decount

        if isinstance(lval, Register):
            target_reg = lval
            load_value = []
        else:
            # not a register, need to put it in one to be able to deref in x86
            target_reg = context.free_registers[0]
            load_value = [
                f"mov {target_reg.full_reg}, {lval.location}",
            ]

        # need to deref decount-1 more times now
        load_value.extend(
            (f"mov {target_reg.full_reg}, [{target_reg.full_reg}]",) * (decount - 1)
        )

        target_address = target_reg.full_reg

        if not isinstance(reg, Register):
            tmp = context.free_registers[1]
            tmp.type = lval.type
            tmp.indirection_count = 0
            load_value.append(f"mov {tmp.location}, {reg}  ; got to be in a reg")
            reg = tmp

        context.mark_free_if_reg(reg)
        context.mark_free_if_reg(target_reg)

        return [
            *asm,
            *get_addr_asm,
            *load_value,
            f"mov {sizespec} [{target_address}], {reg}",  # assign the value
        ], []

    elif (
        node.children[0].type == NodeType.ARRAY_ITEM
    ):  # todo I don't like this specialism branching
        code, (array_item,) = compile(
            node.children[0], context, compile_type=NodeCompilerValueType.lval
        )
        target_address = array_item
        identifier_node, index_expr = node.children[0].children
        # code, (index,) = compile(index_expr, context)
        identifier = context.variables[identifier_node.token.text]
        sizespec = SizeSpecifiers[identifier.size]

        asm = [
            *code,
            f"mov {sizespec} [{target_address}], {reg}   ; {identifier}[...]=... ",  # assign the value
        ]
        context.mark_free_if_reg(reg)
        context.mark_free_if_reg(target_address)
        return asm, []
    else:
        var = context.variables[node.children[0].token.text]
        sizespec = SizeSpecifiers[var.size]
        if var.size != reg.size:
            if isinstance(reg, Literal):
                reg.type = var.type
            else:
                raise CompileError("Size mismatch.", node)

        context.mark_free_if_reg(reg)
        return [*asm, f"mov {sizespec} {var.location}, {reg}"], []


@node_compiler(NodeType.RETURN)
def compile_return(node, context: Context):
    if len(node.children) == 1:
        ret = node.children[0]
        asm, (reg,) = compile(ret, context)
        rax = Register("rax")
        rax.type = reg.type
        rax.indirection_count = reg.indirection_count
        asm = (*asm, "xor rax, rax", f"mov {rax}, {reg}")
        context.mark_free_if_reg(reg)
    else:
        asm = []
    return [*asm, f"jmp {context.current_func.identifier}___ret"], []


@node_compiler(NodeType.TERMINATOR)
def compile_terminator(node, context: Context):
    return [], []


@node_compiler(NodeType.FUNC_CALL)
def compile_call(node, context: Context):
    function_name = node.children[0].token.text
    try:
        function = context.variables[function_name]
        assert isinstance(function, Function)
    except:
        raise CompileError(f"No function {function_name}.", node)

    param_fill_asm = []
    param_push_asm = []
    param_pop_asm = []
    input_params = node.children[1:]
    i = 0
    flt_i = 0
    if function.variadic:
        required_param_count = len(function.parameters) - 1
        extras = [function.parameters[-1]] * (
            len(input_params) - len(function.parameters)
        )
        function_params = function.parameters + extras
    else:
        required_param_count = len(function.parameters)
        function_params = function.parameters

    if len(input_params) < required_param_count:
        raise CompileError(
            f"Function call {function_name} takes {required_param_count} {'or more' if function.variadic else ''} arguments, but {len(input_params)} were given.",
            node,
        )

    for parameter, parameter_decl in zip(input_params, function_params):
        asm, (p_reg,) = compile(parameter, context)
        if isinstance(p_reg, Array):
            # special case, oh no this is getting worse
            # pass as a pointer
            p_reg_ref = context.take_a_register()
            p_reg_ref.type = p_reg.type
            p_reg_ref.indirection_count = 1

            asm.extend([f"lea {p_reg_ref}, {p_reg.location}"])
            context.mark_free_if_reg(p_reg)
            p_reg = p_reg_ref

        if parameter_decl.type != NodeType.ELLIPSIS and (
            p_reg.type != parameter_decl.type
            or p_reg.indirection_count != parameter_decl.indirection_count
        ):
            raise CompileError(
                "Function call parameter type mismatch. "
                + f"'{parameter_decl.identifier}' must be {parameter_decl.type} {'ref' if parameter_decl.indirection_count else ''}"
                + f" but got  {{{p_reg.location}}} : {p_reg.type} {'ref' if p_reg.indirection_count else ''}.",
                node,
            )

        if p_reg.type == VariableType.f64:
            ...
            target_register = FloatArgumentRegisters[flt_i]

            # get the float into memory ...
            if isinstance(p_reg, Register):
                asm.append(f"mov [rsp], {p_reg}")
                target_mem = "[rsp]"
            else:
                target_mem = p_reg
            asm.extend(
                [
                    f"movsd {target_register}, {target_mem}",
                ]
            )
            param_fill_asm.extend(asm)
            flt_i += 1
        else:
            target_register = ArgumentRegistersBySize[p_reg.size][i]
            target_register_full = ArgumentRegistersBySize[8][i]
            asm.extend(
                [
                    f"xor {target_register_full}, {target_register_full}",
                    f"mov {SizeSpecifiers[p_reg.size]} {target_register}, {p_reg}",
                ]
            )
            param_push_asm.append(f"push {target_register_full}")
            param_pop_asm.append(f"pop {target_register_full}")
            param_fill_asm.extend(asm)
            i += 1

        context.mark_free_if_reg(p_reg)

    extra = len(param_push_asm) % 2
    if extra:
        param_push_asm.append("push 0")
        param_pop_asm.append("pop rax")
    param_pop_asm = reversed(param_pop_asm)

    rax = Register("rax")
    rax.type = function.type
    rax.indirection_count = function.indirection_count
    return_val = context.take_a_register()
    return_val.type = function.type
    return_val.indirection_count = function.indirection_count

    return [
        *param_push_asm,
        *param_fill_asm,
        "xor rax, rax",
        f"mov rax, {flt_i}",  # number of float arguments
        f"call {function_name}",
        f"xor {return_val.full_reg}, {return_val.full_reg}",
        f"mov {return_val}, {rax}",
        *param_pop_asm,
    ], [return_val]


@node_compiler(NodeType.FOR_ARRAY_LOOP)
def compile_for(node: Node, context: Context):
    array_id, index_id, element, block = node.children
    loop_id = f"loop_{node.id}"
    array = context.variables.get(array_id.token.text)
    if not array:
        raise CompileError(f"Unknown variable {array_id.token.text}", node)
    if not isinstance(array, Array):
        raise CompileError(
            f"{array_id} is not an array. For loops are for arrays.", node
        )
    context.new_frame()

    # index variable creation
    index = Variable(identifier=index_id.token.text, type=VariableType.u64)
    context.locals_stack_size += index.size
    index.location = RBP(-context.locals_stack_size)
    context.variables[index.identifier] = index

    element = Variable(
        identifier=element.token.text, type=array.type, indirection_count=1
    )
    context.locals_stack_size += element.size
    element.location = RBP(-context.locals_stack_size)
    context.variables[element.identifier] = element

    context.break_out_point.append(f".{loop_id}_end")
    context.continue_point.append(f".{loop_id}_tail")
    loop_body, _ = compile(block, context)
    context.break_out_point.pop()
    context.continue_point.pop()

    asm = [
        f".{loop_id}_begin:",
        f"mov qword {index.location}, 0        ; zero index",
        f"lea rax, {array.location}",
        f"  mov {element.location}, rax",
        f".{loop_id}_start:",
        f"  mov eax, {index.location}",
        f"  cmp eax, {array.length}",
        f"  jge .{loop_id}_end",
        *loop_body,
        f".{loop_id}_tail:",
        f"  add qword {element.location}, {element.base_type_size}",
        f"  add qword {index.location}, 1        ; inc index",
        f"   jmp .{loop_id}_start",
        f".{loop_id}_end:",
    ]

    context.pop_frame()
    return asm, []


@node_compiler(NodeType.WHILE_LOOP)
def compile_while(node: Node, context: Context):
    relational_node, block = node.children
    loop_id = f"while_{node.id}"

    condition_asm, (condition,) = compile(relational_node, context)

    context.break_out_point.append(f".{loop_id}_end")
    context.continue_point.append(f".{loop_id}_start")
    loop_body, _ = compile(block, context)
    context.break_out_point.pop()
    context.continue_point.pop()
    reg_for_comparer = Register("rax")
    reg_for_comparer.type = condition.type
    asm = [
        f".{loop_id}_start:",
        *condition_asm,
        "; check for true",
        f"mov {reg_for_comparer}, {condition.location}",
        f"cmp byte {reg_for_comparer}, 1",
        f"jnz .{loop_id}_end",
        *loop_body,
        f"jmp .{loop_id}_start",
        f".{loop_id}_end:",
    ]
    context.mark_free_if_reg(condition)

    return asm, []


@node_compiler(NodeType.BOOLEAN)
def compile_for_bool(node: Node, context: Context):
    reg = context.take_a_register()
    reg.type = VariableType.u8
    reg.indirection_count = 0

    if node.token.typ == TokenSpec.TRUE:
        asm = [f"mov {reg}, 1"]
    else:
        asm = [f"mov {reg}, 0"]
    return asm, [reg]


@node_compiler(NodeType.FOR_RANGE_LOOP)
def compile_for_range(node: Node, context: Context):
    range_node, identifier_node, block = node.children
    loop_id = f"loop_{node.id}"
    var = typed_identifier_to_variable(identifier_node)
    # todo check the var is an integer...

    # context.new_frame()
    context.locals_stack_size += var.size
    var.location = RBP(-context.locals_stack_size)
    context.variables[var.identifier] = var

    start_asm, (start,) = compile(range_node.children[0], context)
    end_asm, (end,) = compile(range_node.children[1], context)
    step_asm, (step,) = (
        compile(range_node.children[2], context)
        if len(range_node.children) > 2
        else ([], (1,))
    )
    context.break_out_point.append(f".{loop_id}_end")
    context.continue_point.append(f".{loop_id}_tail")
    loop_body, _ = compile(block, context)
    context.break_out_point.pop()
    context.continue_point.pop()
    reg_for_comparer = Register("rax")
    reg_for_comparer.type = var.type
    asm = [
        *start_asm,
        *end_asm,
        *step_asm,
        f".{loop_id}_begin:",
        f"mov {var.sizespec} {var.location}, {start}        ; range start",
        f".{loop_id}_start:",
        f"  mov {reg_for_comparer}, {var.location}",
        f"  cmp {reg_for_comparer}, {end}",
        f"  jge .{loop_id}_end",
        *loop_body,
        f".{loop_id}_tail:",
        f"  add {var.sizespec} {var.location}, {step}        ; inc index",
        f"   jmp .{loop_id}_start",
        f".{loop_id}_end:",
    ]
    context.mark_free_if_reg(start)
    context.mark_free_if_reg(end)
    context.mark_free_if_reg(step)
    # context.pop_frame()
    return asm, []


@node_compiler(NodeType.RELATIONAL)
def compile_relational(node, context: Context):
    left, right = node.children
    left_asm, (left,) = compile(left, context)
    right_asm, (right,) = compile(right, context)

    is_signed = (
        left.type in SignedIntVariableTypes or right.type in SignedIntVariableTypes
    )
    is_float = left.type == VariableType.f64 or right.type == VariableType.f64

    if not is_float:
        condition = node.token.typ
        match condition:
            case TokenSpec.LESS_THAN:
                op = "setb" if not is_signed else "setl"
            case TokenSpec.MORE_THAN:
                op = "seta" if not is_signed else "setg"
            case TokenSpec.LESS_THAN_EQ:
                op = "setbe" if not is_signed else "setle"
            case TokenSpec.MORE_THAN_EQ:
                op = "setae" if not is_signed else "setge"
            case TokenSpec.EQUAL:
                op = "sete"
            case TokenSpec.NOT_EQ:
                op = "setne"

        result = context.take_a_register()
        result.type = VariableType.u8
        result.indirection_count = 0
        rax = Register("rax")
        rax.type = left.type

        asm = [
            *left_asm,
            *right_asm,
            f"mov {rax}, {left}",
            f"cmp {rax}, {right}",  #
            f"{op} al",
            f"mov {result}, al",
        ]
    else:
        condition = node.token.typ
        match condition:
            case TokenSpec.LESS_THAN:
                lreg = "xmm0"
                rreg = "xmm1"
                op = "cmpltsd "
            case TokenSpec.MORE_THAN:
                lreg = "xmm1"
                rreg = "xmm0"
                op = "cmpltsd "
            case TokenSpec.LESS_THAN_EQ:
                lreg = "xmm0"
                rreg = "xmm1"
                op = "cmplesd "
            case TokenSpec.MORE_THAN_EQ:
                lreg = "xmm1"
                rreg = "xmm0"
                op = "cmplesd "
            case TokenSpec.EQUAL:
                lreg = "xmm0"
                rreg = "xmm1"
                op = "cmpeqsd "
            case TokenSpec.NOT_EQ:
                lreg = "xmm0"
                rreg = "xmm1"
                op = "cmpneqsd"

        result = context.take_a_register()
        result.type = VariableType.u8
        result.indirection_count = 0
        asm = [
            *left_asm,
            *right_asm,
        ]
        right = ensure_is_a_register(asm, right, context)
        left = ensure_is_a_register(asm, left, context)
        asm.extend(
            [
                f"mov [rsp-8], {left}",
                "movsd xmm0, [rsp-8]",
                f"mov [rsp-8], {right}",
                "movsd xmm1, [rsp-8]",
                f"{op} {lreg}, {rreg}",
                f"movsd [rsp-8], {lreg}",
                "mov rcx, [rsp-8]",
                "and rcx, 1",
                f"mov byte {result}, cl",
            ]
        )

    context.mark_free_if_reg(left)
    context.mark_free_if_reg(right)
    return asm, [result]


@node_compiler(NodeType.NEGATED_RELATIONAL)
def compile_negated_relational(node, context: Context):
    rel_asm, (relational,) = compile(node.children[0], context)

    if isinstance(relational, Register):
        target_reg = relational
        load_value = []
    else:
        # not a register, need to put it in one to be able to xor it safe
        target_reg = context.take_a_register()
        target_reg.type = relational.type
        target_reg.indirection_count = relational.indirection_count
        load_value = [
            f"mov {target_reg}, {relational.location}",
        ]
        context.mark_free_if_reg(relational)

    return [*rel_asm, *load_value, f"xor {target_reg.sizespec} {target_reg}, 1"], [
        target_reg
    ]


def _compile_and_or_or(node, context: Context, operation):
    result = context.take_a_register()
    result.type = VariableType.u8
    result.indirection_count = 0
    left_asm, (left,) = compile(node.children[0], context)
    right_asm, (right,) = compile(node.children[1], context)
    context.mark_free_if_reg(left)
    context.mark_free_if_reg(right)
    return [
        *left_asm,
        *right_asm,
        f"mov al, {left}",
        f"{operation} al, {right}",
        f"mov {result}, al",
    ], [result]


@node_compiler(parser.NodeType.LOGIC_OR)
def compile_or(node, context: Context) -> list[str]:
    return _compile_and_or_or(node, context, "or")


@node_compiler(parser.NodeType.LOGIC_AND)
def compile_and(node, context: Context) -> list[str]:
    return _compile_and_or_or(node, context, "and")


@node_compiler(parser.NodeType.BITWISE_OP)
def compile_bitop(node, context: Context) -> list[str]:
    result = context.take_a_register()
    op = None
    match node.token.typ:
        case TokenSpec.BIT_NOT:
            op = "not"
        case TokenSpec.BIT_XOR:
            op = "xor"
        case TokenSpec.BIT_OR:
            op = "or"
        case TokenSpec.BIT_AND:
            op = "and"
        case TokenSpec.BIT_LSHIFT:
            op = "shl"
        case TokenSpec.BIT_RSHIFT:
            op = "shr"

    if node.extra.is_binary:
        left, right = node.children
        left_asm, (left,) = compile(left, context)
        right_asm, (right,) = compile(right, context)
        context.mark_free_if_reg(left)
        context.mark_free_if_reg(right)
        result.type = left.type
        result.indirection_count = 0
        asm = [
            *left_asm,
            *right_asm,
            f"mov {result}, {left}",
            f"{op} {result}, {right}",
        ]
    elif node.extra.is_unary:
        left_asm, (left,) = compile(node.children[0], context)
        context.mark_free_if_reg(left)
        result.type = left.type
        result.indirection_count = 0
        asm = [*left_asm, f"mov {result}, {left}", f"{op} {result}"]
    return asm, [result]


@node_compiler(NodeType.BREAK)
def compile_break(node, context: Context):
    if not context.break_out_point:
        raise CompileError("Can't break outside of a loop.", node)
    jmp_to = context.break_out_point[-1]  # dont pop it as it gets popped by the loop
    return [f"jmp {jmp_to}"], []


@node_compiler(NodeType.CONTINUE)
def compile_break(node, context: Context):
    if not context.continue_point:
        raise CompileError("Can't continue outside of a loop.", node)
    jmp_to = context.continue_point[-1]  # dont pop it as it gets popped by the loop
    return [f"jmp {jmp_to}"], []


@node_compiler(NodeType.SQUELCH)
def compile_squelch(node, context: Context):
    target_type, var_node = node.children
    target_type = TokenVariableTypeMap[target_type.token.typ]
    target_size = TypeSizes[target_type]

    prep_var_asm, (var,) = compile(var_node, context)
    if isinstance(var, Literal):
        var.type = target_type
        return [], [var]

    reg = context.take_a_register()
    reg.type = target_type
    reg.indirection_count = var.indirection_count

    source_size = var.size
    if target_size > source_size:
        # squelch up
        reg.type = var.type
    else:
        # squelch down
        reg.type = target_type

    if isinstance(var, Register):
        # could be moving a reg to a reg so need to match
        var.type = reg.type

    asm = [
        *prep_var_asm,
        f"xor {reg.full_reg}, {reg.full_reg}",  # zero it
        f"mov {reg}, {var};",
    ]
    reg.type = target_type
    context.mark_free_if_reg(var)
    return asm, [reg]


@node_compiler(NodeType.IF_STATEMENT)
def compile_if(node, context: Context):
    condition = node.children[0]
    eval_expr, (condition,) = compile(condition, context)

    if isinstance(condition, Literal):
        c = context.free_registers[0]
        c.type = condition.type
        c.indirection_count = condition.indirection_count
        comparison = [f"mov {c}, {condition}", f"cmp byte {c}, 0"]
    else:
        comparison = [
            f"cmp byte {condition}, 0",
        ]

    pos_label = f".pos_{id(node)}"
    end_label = f".end_if_{id(node)}"
    comparison = [
        *comparison,
        f"jnz {pos_label}",
    ]
    context.mark_free_if_reg(condition)

    negative_block, _ = (
        compile_block(node.children[2], context) if len(node.children) > 2 else ([], [])
    )

    positive_block, _ = compile_block(node.children[1], context)

    return [
        *eval_expr,
        *comparison,
        *negative_block,
        f"jmp {end_label}",
        f"{pos_label}:",
        *positive_block,
        f"{end_label}:",
    ], []


def parse_file(file, debug=False):
    with open(file, "r") as in_f:
        src = in_f.read()
    tokens = list(TokenSpec.tokenise(src))
    if debug:
        for p in tokens:
            print(p.typ, p.text)
    module = Module(tokens)
    return module


def compiler(file, debug=False):
    module = parse_file(file, debug)
    precompile(module)

    if not module:
        return None
    print_tree(module)
    return compile(module, Context(filename=file))[0]



from blang.parser import NodeType, Node
from blang.tokeniser import TokenSpec

_node_precompilers = {}


def node_precompiler(type: NodeType):
    def node_precompiler(method):
        def _wrap(node: Node):
            method(node)
            node.extra.precompiled = True

        _node_precompilers[type] = _wrap
        return _wrap

    return node_precompiler


def precompile(node):
    children = [node]
    while children:
        child = children.pop()
        children.extend(child.children)
        if child.type in _node_precompilers:
            _node_precompilers[child.type](child)
            child.extra.precompiled = True
        else:
            print(f"WARNING: No precompiler for {child.type}")


@node_precompiler(NodeType.MODULE)
def precompile_module(node): ...


@node_precompiler(NodeType.BASE_TYPE)
def precompile_base_type(node): ...


@node_precompiler(NodeType.REF_TYPE)
def precompile_ref_type(node): ...


@node_precompiler(NodeType.IDENTIFIER_REF)
def precompile_identifier_ref(node): ...


@node_precompiler(NodeType.ARRAY)
def precompile_array(node): ...


@node_precompiler(NodeType.BITWISE_OP)
def precompile_bitop(node):
    if node.token.type == TokenSpec.BIT_NOT:
        node.extra.is_unary = True
    else:
        node.extra.is_binary = True

    node.extra.produces_value = True

    # check the types here in the new pre-compiler stage....



@node_precompiler(NodeType.IDENTIFIER)
def precompile_identifier(node): ...


@node_precompiler(NodeType.TYPED_IDENTIFER)
def precompile_typed_identifier(node): ...


@node_precompiler(NodeType.DECLARATION)
def precompile_declaration(node): ...


@node_precompiler(NodeType.ASSIGNMENT)
def precompile_assignment(node): ...


@node_precompiler(NodeType.INTEGER)
def precompile_integer(node): ...


@node_precompiler(NodeType.FLOAT)
def precompile_float(node): ...


@node_precompiler(NodeType.RETURN)
def precompile_return(node): ...


@node_precompiler(NodeType.FUNC_CALL)
def precompile_func_call(node): ...


@node_precompiler(NodeType.PARAMETER_LIST)
def precompile_parameter_list(node): ...


@node_precompiler(NodeType.FUNC_DEF)
def precompile_func_def(node): ...


@node_precompiler(NodeType.FUNC_DECL)
def precompile_func_decl(node): ...


@node_precompiler(NodeType.BLOCK)
def precompile_block(node): ...


@node_precompiler(NodeType.CAPTURED_EXPRESSION)
def precompile_captured_expression(node): ...


@node_precompiler(NodeType.DE_REF)
def precompile_de_ref(node): ...


@node_precompiler(NodeType.ARRAY_ITEM)
def precompile_array_item(node): ...


@node_precompiler(NodeType.BOOLEAN)
def precompile_boolean(node): ...


@node_precompiler(NodeType.ADDITIVE)
def precompile_additive(node): ...


@node_precompiler(NodeType.TERM)
def precompile_term(node): ...


@node_precompiler(NodeType.RELATIONAL)
def precompile_relational(node): ...


@node_precompiler(NodeType.LOGIC_AND)
def precompile_logic_and(node): ...


@node_precompiler(NodeType.LOGIC_OR)
def precompile_logic_or(node): ...


@node_precompiler(NodeType.NEGATED_RELATIONAL)
def precompile_negated_relational(node): ...


@node_precompiler(NodeType.IF_STATEMENT)
def precompile_if_statement(node): ...


@node_precompiler(NodeType.WHILE_LOOP)
def precompile_while_loop(node): ...


@node_precompiler(NodeType.BREAK)
def precompile_break(node): ...


@node_precompiler(NodeType.CONTINUE)
def precompile_continue(node): ...


@node_precompiler(NodeType.RANGE)
def precompile_range(node): ...


@node_precompiler(NodeType.FOR_ARRAY_LOOP)
def precompile_for_array_loop(node): ...


@node_precompiler(NodeType.FOR_RANGE_LOOP)
def precompile_for_range_loop(node): ...


@node_precompiler(NodeType.SQUELCH)
def precompile_squelch(node): ...


@node_precompiler(NodeType.STRING)
def precompile_string(node): ...


@node_precompiler(NodeType.CHARACTER)
def precompile_character(node): ...


@node_precompiler(NodeType.TERMINATOR)
def precompile_terminator(node): ...


@node_precompiler(NodeType.IMPORT)
def precompile_import(node): ...



from collections import ChainMap
from dataclasses import dataclass, field
import enum
from blang.parser import TokenSpec, Node
from pathlib import Path


class VariableType(enum.StrEnum):
    u8 = enum.auto()
    u16 = enum.auto()
    u32 = enum.auto()
    u64 = enum.auto()

    i8 = enum.auto()
    i16 = enum.auto()
    i32 = enum.auto()
    i64 = enum.auto()

    f64 = enum.auto()


SignedIntVariableTypes = [
    VariableType.i8,
    VariableType.i16,
    VariableType.i32,
    VariableType.i64,
]

TokenVariableTypeMap = {
    TokenSpec.U8: VariableType.u8,
    TokenSpec.U16: VariableType.u16,
    TokenSpec.U32: VariableType.u32,
    TokenSpec.U64: VariableType.u64,
    TokenSpec.I8: VariableType.i8,
    TokenSpec.I16: VariableType.i16,
    TokenSpec.I32: VariableType.i32,
    TokenSpec.I64: VariableType.i64,
    TokenSpec.F64: VariableType.f64,
}

TypeSizes = {
    VariableType.u8: 1,
    VariableType.u16: 2,
    VariableType.u32: 4,
    VariableType.u64: 8,
    VariableType.i8: 1,
    VariableType.i16: 2,
    VariableType.i32: 4,
    VariableType.i64: 8,
    VariableType.f64: 8,
}

SizeReserves = {1: "resb", 2: "resw", 4: "resd", 8: "resq", 16: "resdq"}

SizeDefiners = {1: "db", 2: "dw", 4: "dd", 8: "dq", 16: "do"}
for t, s in TypeSizes.items():
    SizeDefiners[t] = SizeDefiners[s]

SizeSpecifiers = {1: "byte", 2: "word", 4: "dword", 8: "qword", 16: "oword"}
# for t, s in TypeSizes.items():
# SizeSpecifiers[t] = SizeSpecifiers[s]

ArgumentRegistersBySize = {
    8: ["rdi", "rsi", "rdx", "rcx", "r8", "r9"],
    4: ["edi", "esi", "edx", "ecx", "r8d", "r9d"],
    2: ["di", "si", "dx", "cx", "r8w", "r9w"],
    1: ["dil", "sil", "dxl", "cxl", "r8b", "r9b"],
}

FloatArgumentRegisters = [f"xmm{i}" for i in range(7)]

# GP Registers split by size
GPRegisters = ("r10", "r11", "r12", "r13", "r14", "r15")
RegistersPartial = {
    r: {1: f"{r}b", 2: f"{r}w", 4: f"{r}d", 8: f"{r}"} for r in GPRegisters
}
RegistersPartial.update(
    {
        "rax": {8: "rax", 4: "eax", 2: "ax", 1: "al"},
        "rbp": {8: "rbp"},
        "rcx": {8: "rcx", 4: "ecx", 2: "cx", 1: "cl"},
        "rdx": {8: "rdx", 4: "edx", 2: "dx", 1: "dl"},
    }
)


class SizePropMixin:
    @property
    def size(self) -> int:
        return 8 if self.indirection_count else self.base_type_size

    @property
    def base_type_size(self) -> int:
        return TypeSizes[self.type]

    @property
    def sizespec(self):
        return SizeSpecifiers[self.size]


@dataclass
class Register(SizePropMixin):
    full_reg: str

    # Similar to a variable
    type: VariableType = None
    indirection_count: int = 0

    @property
    def content_type_size(self):
        return self.content_type and TypeSizes[self.content_type]

    @property
    def name(self):
        return RegistersPartial[self.full_reg][self.size]

    def partial(self, size):
        return RegistersPartial[self.full_reg][size]

    @property
    def location(self):
        return self.name

    def set_in_use(self, size):
        pass

    def __str__(self):
        return self.name or self.full_reg


@dataclass
class Literal(SizePropMixin):
    value: str

    # Similar to a variable
    type: VariableType = None
    indirection_count: int = 0
    length: int = 1

    def __str__(self):
        return self.value


@dataclass
class Variable(SizePropMixin):
    identifier: str
    type: VariableType

    location: str | None = None
    on_stack: bool = False
    indirection_count: int = 0
    external: bool = False
    exported: bool = False
    node: Node = None

    def __str__(self):
        return str(self.location)


@dataclass
class Array(SizePropMixin):
    identifier: str
    type: VariableType

    location: str | None = None
    on_stack: bool = False
    indirection_count: int = 0
    external: bool = False
    exported: bool = False
    node: Node = None
    length: int = 1  # Number of elements if an array

    def __str__(self):
        return str(self.location)


class LiteralString(Literal, Array):
    def __init__(self, *args, **kwargs):
        val = kwargs.pop("value")
        Literal.__init__(self, val)
        Array.__init__(self, *args, **kwargs)


class NodeCompilerValueType(enum.Enum):
    lval = enum.auto()
    rval = enum.auto()


@dataclass
class AddressPointer:
    # in a register
    base: Register | str
    offset: int = 0

    def __str__(self):
        return f"[{self.address}]"

    @property
    def address(self):
        return f"{self.base}{f' + {self.offset}' if self.offset > 0 else f' - {abs(self.offset)}' if self.offset < 0 else ''}"


def test_ap():
    a = RBP(-9)
    print(f"{a}")
    a = RBP(9)
    print(f"{a}")
    a = RBP(0)
    print(f"{a}")
    assert False


def RBP(offset):
    return AddressPointer("rbp", offset)


@dataclass
class Function(Variable):
    parameters: list[Variable] = field(default_factory=list)
    variadic: bool = False


class Context: ...


@dataclass
class FloatRegister:
    name: str
    context: Context
    size: int = 8


@dataclass
class Context:
    variable_stack: ChainMap = field(default_factory=ChainMap)
    string_literals: dict[str, str] = field(default_factory=dict)
    float_literals: dict[str, Variable] = field(default_factory=dict)
    locals_stack_size: int = 0
    use_stack: bool = False
    current_func: Function = None
    free_registers: list[Register] = field(
        default_factory=lambda: [
            Register("r10"),
            Register("r11"),
            Register("r12"),
            Register("r13"),
            Register("r14"),
            Register("r15"),
        ]
    )
    occupied_registers: list[Register] = field(default_factory=lambda: [])

    free_float_registers: list[str] = field(
        default_factory=lambda: [f"xmm{i}" for i in range(15)]
    )
    occupied_float_registers: list[str] = field(default_factory=lambda: [])

    break_out_point: list[str] = field(default_factory=lambda: [])
    continue_point: list[str] = field(default_factory=lambda: [])
    strings: dict = field(default_factory=dict)
    filename: Path | None = None

    init_asm = []

    @property
    def globals_(self):
        return self.variable_stack.maps[-1]

    @property
    def variables(self):
        return self.variable_stack

    def is_local_var(self, var):
        return var in self.variable_stack.maps[0]

    def pop_frame(self):
        self.variable_stack = self.variable_stack.parents

    def new_frame(self):
        self.variable_stack = self.variable_stack.new_child()

    def take_a_register(self):
        reg = self.free_registers.pop()
        self.occupied_registers.append(reg)
        return reg

    def mark_free_if_reg(self, maybe_reg):
        if isinstance(maybe_reg, Register):
            if maybe_reg in self.occupied_registers:
                self.occupied_registers.remove(maybe_reg)
            if maybe_reg.full_reg in GPRegisters:
                self.free_registers.append(maybe_reg)

    def take_a_float_register(self):
        reg = self.free_float_registers.pop()
        self.occupied_float_registers.append(reg)
        return reg

    def free_a_float_register(self, reg):
        self.occupied_float_registers.remove(reg)
        self.free_float_registers.append(reg)


def test_var_stack():
    c = Context()
    c.globals_["a"] = 1
    c.new_frame()
    c._locals_["a"] = 5
